#include <malloc.h>
#include <mpi.h>
#include <mpe.h>
#include "tbx_timer.h"

#ifndef MACHINE_H
#define MACHINE_H

struct group_info
{
   MPI_Group  group;
   MPI_Comm   group_comm;
   int  number_of_tasks;
   int *tasks_list;
};

typedef struct group_info group_info;

struct global_info
{
   int  my_address;

   group_info global;
   group_info node;
};

typedef struct global_info global_info;

struct mcb
{
   MPI_Request request;
   int node;
   int source;
   int dest;
   int tag;
   int bytes;
   char *buffer;
   void (*handler)();
   void *handler_arg;
};

typedef struct mcb * CMMD_mcb;

#ifdef MACHINE_MAIN
global_info machine_info;
int Host;
int msgid;
tbx_timer Timers[64];
#else
extern global_info machine_info;
extern int Host;
extern int msgid;
extern tbx_timer Timers[64];
#endif

#define CMMD_ANY_NODE MPI_ANY_SOURCE
#define CMMD_ANY_TAG  MPI_ANY_TAG

#define enable_diag_write();
#define CMMD_fset_io_mode(a,b);

/* Host communication */
#define CMMD_enable();

#define CMMD_host_node() Host


/* Node communication */
#define CMMD_enable_host();

#define CMMD_self_address() machine_info.my_address
#define CMMD_partition_size() machine_info.node.number_of_tasks


/* Comminication maching pairs */
#define CMMD_send(destination, tag, buffer, len) \
        {                                                                     \
           MPE_Log_send(destination, tag, len);                               \
           MPE_Log_event(1,0,"Send");                                         \
           MPI_Send(buffer, len, MPI_BYTE, destination, tag, MPI_COMM_WORLD); \
           MPE_Log_event(2,0,"Send");                                         \
        }

#define CMMD_receive(source, tag, buffer, len)                          \
        {                                                               \
           MPI_Status status;                                           \
                                                                        \
           MPE_Log_event(3,0,"Receive");                                \
           MPI_Recv(buffer, len, MPI_BYTE, source, tag, MPI_COMM_WORLD, \
                    &status);                                           \
           MPE_Log_event(4,0,"Receive");                                \
           MPE_Log_receive(status.MPI_SOURCE, status.MPI_TAG, len);     \
        }

#define CMMD_bc_from_host(buffer, len) \
        MPI_Bcast(buffer, len, MPI_BYTE, Host, machine_info.global.group_comm)

#define CMMD_receive_bc_from_host(buffer, len)                  \
        MPI_Bcast(buffer, len, MPI_BYTE, Host, machine_info.global.group_comm)

#define CMMD_error printf


/* Sync. */

#define CMMD_sync_host_with_nodes()                               \
        {                                                         \
           MPE_Log_event(5,1,"Before CMMD_sync_host_with_nodes"); \
           MPI_Barrier(machine_info.global.group_comm);           \
           MPE_Log_event(6,1,"After CMMD_sync_host_with_nodes");  \
        }
#define CMMD_sync_with_host()                               \
        {                                                   \
           MPE_Log_event(5,1,"Before CMMD_sync_with_host"); \
           MPI_Barrier(machine_info.global.group_comm);     \
           MPE_Log_event(6,1,"After CMMD_sync_with_host");  \
        }
#define CMMD_sync_with_nodes()                               \
        {                                                    \
           MPE_Log_event(5,2,"Before CMMD_sync_with_nodes"); \
           MPI_Barrier(machine_info.node.group_comm);        \
           MPE_Log_event(6,2,"After CMMD_sync_with_nodes");  \
        }

/* ASync. communication */

#define  CMMD_mcb_node(mcb)     (mcb) -> node
#define  CMMD_mcb_source(mcb)   (mcb) -> source
#define  CMMD_mcb_dest(mcb)     (mcb) -> dest
#define  CMMD_mcb_tag(mcb)      (mcb) -> tag
#define  CMMD_mcb_bytes(mcb)    (mcb) -> bytes
#define  CMMD_mcb_buffer(mcb)   (mcb) -> buffer

#define CMMD_free_mcb(mcb)     free(mcb)

/* Timing */
#define CMMD_node_timer_clear(a) tbx_timer_reset(&(Timers[a]))
#define CMMD_node_timer_start(a) tbx_timer_start(&(Timers[a]))
#define CMMD_node_timer_stop(a)         \
        {                               \
           tbx_timer_stop(&(Timers[a])); \
           compute_time(&(Timers[a]));   \
        }
#define CMMD_node_timer_elapsed(a)                                   \
        (double) Timers[a].total_sec + (double) Timers[a].total_usec / 1000000.0
#define CMMD_node_timer_busy(a)                                      \
        (double) Timers[a].total_sec + (double) Timers[a].total_usec / 1000000.0
#define CMMD_node_timer_idle(a)  0.0

#endif

CMMD_mcb CMMD_send_async(int dest_node, int tag, void *buffer, int buffer_desc,
                         void (*handler)(CMMD_mcb *, void *),
                         void *handler_arg);
CMMD_mcb CMMD_receive_async(int source_node, int tag, void *buffer,
                            int buffer_desc,
                            void (*handler)(CMMD_mcb *, void *),
                            void *handler_arg);
void CMMD_msg_wait(CMMD_mcb mcb);


